import math

from ngsolve.comp import Mesh
from netgen.geom2d import unit_square
from ngsolve import CoefficientFunction, Trace, Id, SetNumThreads, InnerProduct, dx, BND, SymbolicBFI, Draw, \
    SetVisualization
from ngsolve.comp import H1, GridFunction, BilinearForm
from ngsolve.fem import Parameter

ITER = 50
STEPS = 1
ERROR = 1e-10


def getStressTensor(u, mu, lam, mesh):
    return 2 * mu * getEpsilonTensor(u) + lam * Trace(getEpsilonTensor(u)) * Id(mesh.dim)


def getEpsilonTensor(u):
    return 0.5 * (u.Deriv().trans + u.Deriv())


geometry = unit_square
geometry.AppendPoint(0, 1, name='point', maxh=0.03)
geometry.SetMaterial(1, 'material')

mesh = Mesh(geometry.GenerateMesh(maxh=0.5))

print('\n')
print('Edge bc names:', mesh.GetBoundaries())
print('Point bc names:', mesh.GetBBoundaries())
print('Materials', mesh.GetMaterials())
print('\n')

E, nu = 10.0, 0.3
mu = E / 2./(1+nu)
lmbda = 2 * mu * nu/(1-2*nu)

force = CoefficientFunction((-0.5, 0))
functionMu = CoefficientFunction([mu])
functionLam = CoefficientFunction([lmbda])

conformingH1 = H1(mesh, order=3, dirichlet='left', dim=mesh.dim)
function = GridFunction(conformingH1)
trial, test = conformingH1.TrialFunction(), conformingH1.TestFunction()
inner = InnerProduct(getEpsilonTensor(trial), getEpsilonTensor(test))
trace = Trace(trial.Deriv()) * Trace(test.Deriv())

linear = BilinearForm(conformingH1)
linear += (2 * functionMu * inner + functionLam * trace) * dx
linear += SymbolicBFI(- Parameter(1) * force * test, definedon=mesh.BBoundaries('point'))

result = function.vec.CreateVector()
inversion = function.vec.CreateVector()

for step in range(1, STEPS + 1):
    for it in range(1, ITER + 1):

        linear.Apply(function.vec, result)
        linear.AssembleLinearization(function.vec)

        inv = linear.mat.Inverse(conformingH1.FreeDofs(), inverse='umfpack')
        inversion.data = inv * result
        function.vec.data -= inversion
        del inv

        error = math.sqrt(abs(InnerProduct(inversion, result)))

        if error < ERROR:
            break

        Draw(function, mesh, "displacement")
        SetVisualization(deformation=False)
